#ifndef SYMBOL_H_
#define SYMBOL_H_

#include "common.h"
#include "as_tree.h"

/* Symbol specifiers. */
typedef enum {
	SYMBOL_TYPE_UNKNOWN,
	SYMBOL_TYPE_VARIABLE,
	SYMBOL_TYPE_ARGUMENT,
	SYMBOL_TYPE_ARRAY,
	SYMBOL_TYPE_FUNC,
	SYMBOL_TYPE_FUNC_DECL
} sym_type_t;

typedef enum {
	SYMBOL_VISIB_UKNOWN,
	SYMBOL_VISIB_GLOBAL,
	SYMBOL_VISIB_MODULE,
	SYMBOL_VISIB_EXTERN
} storage_class_t;

struct symbol_table {
	struct symbol_table *prev;
	struct hash_table *scope;
	unsigned int nargs;
	unsigned int nvars;
	unsigned int ref_count;
};

struct symbol;
typedef void (*symbol_release_t)(struct symbol * );

struct frame_pos {
	int		pos;
	unsigned int	size;
};

struct symbol {
	sym_type_t	type;
	unsigned int	ref_count;
	data_type_t	decl_type;
	storage_class_t	klass;
	char		*name;
	const char	*file_name;
	unsigned int	line_number;
	symbol_release_t release;
	unsigned int	nr_references;
	struct frame_pos fr;
};

struct symbol_var {
	struct symbol	base;
	unsigned int	scope_num;
};

struct symbol_arg {
	struct symbol_var base;
	struct ast_node *initval;
	int		is_fmt :1;
	int		is_ref :1;
};

struct symbol_array {
	struct symbol_var base;
	unsigned int	*dims;
	unsigned int	ndims;
};

struct symbol_func {
	struct symbol	base;
	unsigned int	with_fmt :1;
	struct symbol_arg **args;
	unsigned int	nargs;
};

#define	SYMBOL(klass)		((struct symbol *)(klass))
#define	SYMBOL_FUNC(klass)	((struct symbol_func *)(klass))
#define	SYMBOL_VAR(klass)	((struct symbol_var *)(klass))
#define	SYMBOL_ARG(klass)	((struct symbol_arg *)(klass))
#define	SYMBOL_ARRAY(klass)	((struct symbol_array *)(klass))

struct symbol_var *symbol_variable(const char *name, data_type_t decl_type, storage_class_t klass);
void symbol_variable_set_initval(struct symbol_var *symbol, struct ast_node *initval);

struct symbol_array *symbol_array(const char *name, data_type_t decl_type, storage_class_t klass);
void	symbol_array_add_dim(struct symbol_array *symbol, unsigned int size);
//void	symbol_array_set_initval(struct symbol_array *symbol, struct ast_node *initval);

struct symbol_func *symbol_func(const char *name, data_type_t ret_type, storage_class_t klass);
void symbol_func_add_arg(struct symbol_func *func, struct symbol_arg *arg);
void symbol_func_mark_as_decl(struct symbol_func *func);

struct symbol_arg *symbol_argument();
void symbol_arg_set_name(struct symbol_arg *symbol, const char *name);
void symbol_arg_set_type(struct symbol_arg *symbol, data_type_t type);
void symbol_arg_set_initval(struct symbol_arg *symbol, struct ast_node *initval);

void symbol_table_put_symbol(struct symbol *symbol);
void symbol_table_remove_symbol(const char *name);

struct symbol **symbol_table_extract_variables(struct symbol_table *table, unsigned int *size);
struct symbol **symbol_table_extract_arguments(struct symbol_table *table, unsigned int *size);

void  symbol_incr_reference(struct symbol *symbol);

void symbol_ref(struct symbol *symbol);
void symbol_unref(struct symbol *symbol);
void symbol_free(struct symbol *symbol);

/* Symbol table related functions. */
void symbol_table_push();
void symbol_table_pop();
void symbol_table_put(struct symbol_table *table);
void symbol_table_create_global_table();

void symbol_table_ref(struct symbol_table *table);
void symbol_table_unref(struct symbol_table *table);

struct symbol_table *symbol_table_get_current_table();
struct symbol_table *symbol_table_get_global_table();

struct symbol *symbol_table_lookup_top(const char *name);
struct symbol *symbol_table_lookup_all(const char *name);

#endif /*SYMBOL_H_*/
